# -*- coding: utf-8 -*-

import math

from transformers.models.t5.modeling_t5 import *
from transformers.models.t5.modeling_t5 import _CONFIG_FOR_DOC


@add_start_docstrings("""T5 Model with a `language modeling` head on top.""", T5_START_DOCSTRING)
class T5ForGenerativeDeduplication(T5PreTrainedModel):
    _keys_to_ignore_on_load_unexpected = [
        "decoder.block.0.layer.1.EncDecAttention.relative_attention_bias.weight",
    ]
    _tied_weights_keys = ["encoder.embed_tokens.weight", "decoder.embed_tokens.weight", "lm_head.weight"]

    def __init__(self, config: T5Config):
        super().__init__(config)
        self.model_dim = config.d_model

        self.shared = nn.Embedding(config.vocab_size, config.d_model)

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False
        self.encoder = T5Stack(encoder_config, self.shared)

        decoder_config = copy.deepcopy(config)
        decoder_config.is_decoder = True
        decoder_config.is_encoder_decoder = False
        decoder_config.num_layers = config.num_decoder_layers
        self.decoder = T5Stack(decoder_config, self.shared)
        
        self.lm_head = nn.Linear(config.d_model, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

        # Model parallel
        self.model_parallel = False
        self.device_map = None
        
        self.gd_setup_flag = False
        self.gaussian_noise_prob = None
        self.gaussian_noise_variance = None

    def setup_gd(self,
                 gaussian_noise_prob: float = 0.3,
                 gaussian_noise_variance: float = 0.1):
        self.gd_setup_flag = True
        self.gaussian_noise_prob = gaussian_noise_prob
        self.gaussian_noise_variance = gaussian_noise_variance

    @add_start_docstrings(PARALLELIZE_DOCSTRING)
    def parallelize(self, device_map=None):
        warnings.warn(
            "`T5ForGenerativeDeduplication.parallelize` is deprecated and will be removed in v5 of Transformers, you"
            " should load your model with `device_map='balanced'` in the call to `from_pretrained`. You can also"
            " provide your own `device_map` but it needs to be a dictionary module_name to device, so for instance"
            " {'encoder.block.0': 0, 'encoder.block.1': 1, ...}",
            FutureWarning,
        )
        self.device_map = (
            get_device_map(len(self.encoder.block), range(torch.cuda.device_count()))
            if device_map is None
            else device_map
        )
        assert_device_map(self.device_map, len(self.encoder.block))
        self.encoder.parallelize(self.device_map)
        self.decoder.parallelize(self.device_map)
        self.lm_head = self.lm_head.to(self.decoder.first_device)
        self.model_parallel = True

    @add_start_docstrings(DEPARALLELIZE_DOCSTRING)
    def deparallelize(self):
        warnings.warn(
            "Like `parallelize`, `deparallelize` is deprecated and will be removed in v5 of Transformers.",
            FutureWarning,
        )
        self.encoder.deparallelize()
        self.decoder.deparallelize()
        self.encoder = self.encoder.to("cpu")
        self.decoder = self.decoder.to("cpu")
        self.lm_head = self.lm_head.to("cpu")
        self.model_parallel = False
        self.device_map = None
        torch.cuda.empty_cache()

    def get_input_embeddings(self):
        return self.shared

    def set_input_embeddings(self, new_embeddings):
        self.shared = new_embeddings
        self.encoder.set_input_embeddings(new_embeddings)
        self.decoder.set_input_embeddings(new_embeddings)

    def _tie_weights(self):
        if self.config.tie_word_embeddings:
            self._tie_or_clone_weights(self.encoder.embed_tokens, self.shared)
            self._tie_or_clone_weights(self.decoder.embed_tokens, self.shared)

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    def get_output_embeddings(self):
        return self.lm_head

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    @add_start_docstrings_to_model_forward(T5_INPUTS_DOCSTRING)
    @replace_return_docstrings(output_type=Seq2SeqLMOutput, config_class=_CONFIG_FOR_DOC)
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.BoolTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        decoder_head_mask: Optional[torch.FloatTensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        decoder_inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.FloatTensor], Seq2SeqLMOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[-100, 0, ...,
            config.vocab_size - 1]`. All labels set to `-100` are ignored (masked), the loss is only computed for
            labels in `[0, ..., config.vocab_size]`

        Returns:

        Examples:

        ```python
        >>> from transformers import AutoTokenizer, T5ForGenerativeDeduplication

        >>> tokenizer = AutoTokenizer.from_pretrained("t5-small")
        >>> model = T5ForGenerativeDeduplication.from_pretrained("t5-small")

        >>> # training
        >>> input_ids = tokenizer("The <extra_id_0> walks in <extra_id_1> park", return_tensors="pt").input_ids
        >>> labels = tokenizer("<extra_id_0> cute dog <extra_id_1> the <extra_id_2>", return_tensors="pt").input_ids
        >>> outputs = model(input_ids=input_ids, labels=labels)
        >>> loss = outputs.loss
        >>> logits = outputs.logits

        >>> # inference
        >>> input_ids = tokenizer(
        ...     "summarize: studies have shown that owning a dog is good for you", return_tensors="pt"
        ... ).input_ids  # Batch size 1
        >>> outputs = model.generate(input_ids)
        >>> print(tokenizer.decode(outputs[0], skip_special_tokens=True))
        >>> # studies have shown that owning a dog is good for you.
        ```"""
        if self.training:
            assert self.gd_setup_flag, 'For training, please setup_gd() first!'

        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        # FutureWarning: head_mask was separated into two input args - head_mask, decoder_head_mask
        if head_mask is not None and decoder_head_mask is None:
            if self.config.num_layers == self.config.num_decoder_layers:
                warnings.warn(__HEAD_MASK_WARNING_MSG, FutureWarning)
                decoder_head_mask = head_mask

        # Encode if needed (training, first prediction pass)
        if encoder_outputs is None:
            # Convert encoder inputs in embeddings if needed
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                head_mask=head_mask,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )
        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        hidden_states = encoder_outputs[0]

        # add noise
        if self.training and self.gaussian_noise_prob > 0:
            noise_shape = hidden_states[:, :, :1].shape
            noise = (self.gaussian_noise_variance**0.5) * torch.randn(noise_shape)  # variance 0.1
            noise = noise.to(hidden_states.device)
            # hidden_states = nn.Dropout(0.2)(hidden_states) + noise
            # hidden_states += noise
            noise_mask_binomial = torch.distributions.binomial.Binomial(
                probs=self.gaussian_noise_prob)
            noise_mask = noise_mask_binomial.sample(noise_shape)
            noise_mask = noise_mask.to(hidden_states.device)
            hidden_states = hidden_states * (1.0 - noise_mask) * (1.0 / (1.0 - self.gaussian_noise_prob))
            # print('hidden_states>>>', hidden_states)
            hidden_states += noise_mask * noise
            # add noise end

        if self.model_parallel:
            torch.cuda.set_device(self.decoder.first_device)

        if labels is not None and decoder_input_ids is None and decoder_inputs_embeds is None:
            # get decoder inputs from shifting lm labels to the right
            decoder_input_ids = self._shift_right(labels)

        # Set device for model parallelism
        if self.model_parallel:
            torch.cuda.set_device(self.decoder.first_device)
            hidden_states = hidden_states.to(self.decoder.first_device)
            if decoder_input_ids is not None:
                decoder_input_ids = decoder_input_ids.to(self.decoder.first_device)
            if attention_mask is not None:
                attention_mask = attention_mask.to(self.decoder.first_device)
            if decoder_attention_mask is not None:
                decoder_attention_mask = decoder_attention_mask.to(self.decoder.first_device)

        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            inputs_embeds=decoder_inputs_embeds,
            past_key_values=past_key_values,
            encoder_hidden_states=hidden_states,
            encoder_attention_mask=attention_mask,
            head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        sequence_output = decoder_outputs[0]

        # Set device for model parallelism
        if self.model_parallel:
            torch.cuda.set_device(self.encoder.first_device)
            self.lm_head = self.lm_head.to(self.encoder.first_device)
            sequence_output = sequence_output.to(self.lm_head.weight.device)

        if self.config.tie_word_embeddings:
            # Rescale output before projecting on vocab
            # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586
            sequence_output = sequence_output * (self.model_dim**-0.5)

        lm_logits = self.lm_head(sequence_output)

        loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss(ignore_index=-100)
            # move labels to correct device to enable PP
            labels = labels.to(lm_logits.device)
            loss = loss_fct(lm_logits.view(-1, lm_logits.size(-1)), labels.view(-1))
            if math.isnan(loss.item()):
                print('Oh no! Got a NaN!')

        if not return_dict:
            output = (lm_logits,) + decoder_outputs[1:] + encoder_outputs
            return ((loss,) + output) if loss is not None else output

        return Seq2SeqLMOutput(
            loss=loss,
            logits=lm_logits,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )

    def prepare_inputs_for_generation(
        self,
        input_ids,
        past_key_values=None,
        attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        decoder_attention_mask=None,
        cross_attn_head_mask=None,
        use_cache=None,
        encoder_outputs=None,
        **kwargs,
    ):
        # cut decoder_input_ids if past_key_values is used
        if past_key_values is not None:
            past_length = past_key_values[0][0].shape[2]

            # Some generation methods already pass only the last input ID
            if input_ids.shape[1] > past_length:
                remove_prefix_length = past_length
            else:
                # Default to old behavior: keep only final ID
                remove_prefix_length = input_ids.shape[1] - 1

            input_ids = input_ids[:, remove_prefix_length:]

        return {
            "decoder_input_ids": input_ids,
            "past_key_values": past_key_values,
            "encoder_outputs": encoder_outputs,
            "attention_mask": attention_mask,
            "head_mask": head_mask,
            "decoder_head_mask": decoder_head_mask,
            "decoder_attention_mask": decoder_attention_mask,
            "cross_attn_head_mask": cross_attn_head_mask,
            "use_cache": use_cache,
        }

    def prepare_decoder_input_ids_from_labels(self, labels: torch.Tensor):
        return self._shift_right(labels)

    def _reorder_cache(self, past_key_values, beam_idx):
        # if decoder past is not included in output
        # speedy decoding is disabled and no need to reorder
        if past_key_values is None:
            logger.warning("You might want to consider setting `use_cache=True` to speed up decoding")
            return past_key_values

        reordered_decoder_past = ()
        for layer_past_states in past_key_values:
            # get the correct batch idx from layer past batch dim
            # batch dim of `past` is at 2nd position
            reordered_layer_past_states = ()
            for layer_past_state in layer_past_states:
                # need to set correct `past` for each of the four key / value states
                reordered_layer_past_states = reordered_layer_past_states + (
                    layer_past_state.index_select(0, beam_idx.to(layer_past_state.device)),
                )

            if reordered_layer_past_states[0].shape != layer_past_states[0].shape:
                raise ValueError(
                    f"reordered_layer_past_states[0] shape {reordered_layer_past_states[0].shape} and layer_past_states[0] shape {layer_past_states[0].shape} mismatched"
                )
            if len(reordered_layer_past_states) != len(layer_past_states):
                raise ValueError(
                    f"length of reordered_layer_past_states {len(reordered_layer_past_states)} and length of layer_past_states {len(layer_past_states)} mismatched"
                )

            reordered_decoder_past = reordered_decoder_past + (reordered_layer_past_states,)
        return reordered_decoder_past



# -*- coding: utf-8 -*-

import os
import json
from typing import Optional, Dict
from collections import defaultdict

import torch
from tqdm import tqdm
from datasets import Dataset
from transformers import (
    AutoTokenizer, DataCollatorForSeq2Seq,
    Seq2SeqTrainingArguments, Seq2SeqTrainer
)

from .models.modeling_t5 import T5ForGenerativeDeduplication


ALL_SUPPORT_GEN_MODELS = {
    'T5ForGenerativeDeduplication': T5ForGenerativeDeduplication,
}


class GenDedup:
    def __init__(self,
                 model_name_or_path: str,
                 model_class_name: str = 'T5ForGenerativeDeduplication',
                 pretrained_model_name_or_path: Optional[str] = None,
                 max_length: Optional[int] = None,
                 gaussian_noise_prob: float = 0.1,
                 gaussian_noise_variance: float = 0.1):
        model_class = ALL_SUPPORT_GEN_MODELS[model_class_name]
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        self.model = model_class.from_pretrained(pretrained_model_name_or_path or model_name_or_path)
        self.model.setup_gd(
            gaussian_noise_prob=gaussian_noise_prob,
            gaussian_noise_variance=gaussian_noise_variance)
        self.max_length = max_length
        self.ds = None
        self.feature_columns = self.tokenizer('gen dedup').keys()

    def prepare_ds(self, ds: Dataset) -> Dataset:
        def preprocess_function(examples):
            inputs = [d for d in examples["sentence"]]
            targets = examples['labels']
            model_inputs = self.tokenizer(inputs, max_length=self.max_length, truncation=True)
            labels = self.tokenizer(text_target=targets, max_length=self.max_length, truncation=True)
            model_inputs["labels"] = labels["input_ids"]
            model_inputs["label_text"] = examples['labels']
            return model_inputs

        return ds.map(preprocess_function, batched=True)

    def fit(self,
            ds: Dataset,
            output_dir: str,
            batch_size: int = 32,
            epochs: int = 1,
            learning_rate: float = 1e-4,
            warmup_steps: int = 0,
            logging_steps: int = 10,
            weight_decay: float = 0.01,
            gradient_accumulation_steps: int = 1,
            fp16: Optional[bool] = None,
            argument_kwargs: Optional[Dict] = None,
            trainer_kwargs: Optional[Dict] = None):
        self.ds = self.prepare_ds(ds)
        data_collator = DataCollatorForSeq2Seq(tokenizer=self.tokenizer, model=self.model)

        if argument_kwargs is None:
            argument_kwargs = {}
        if trainer_kwargs is None:
            trainer_kwargs = {}
        training_args = Seq2SeqTrainingArguments(
            output_dir=output_dir,
            learning_rate=learning_rate,
            per_device_train_batch_size=batch_size,
            weight_decay=weight_decay,
            save_total_limit=1,
            num_train_epochs=epochs,
            logging_steps=logging_steps,
            warmup_steps=warmup_steps,
            gradient_accumulation_steps=gradient_accumulation_steps,
            fp16=fp16,
            predict_with_generate=True,
            prediction_loss_only=True,
            **argument_kwargs
        )

        trainer = Seq2SeqTrainer(
            model=self.model,
            args=training_args,
            train_dataset=self.ds,
            tokenizer=self.tokenizer,
            data_collator=data_collator,
        )

        trainer.train()
        self.model.eval()

    def deduplicate(self,
                    save_dir: str,
                    max_label_words: int = 1,
                    ds: Optional[Dataset] = None,
                    threshold: float = 0.5,
                    device: Optional[str] = None,
                    generate_kwargs: Optional[Dict] = None,
                    exist_ok: bool = True) -> Dict:
        if generate_kwargs is None:
            generate_kwargs = {}

        os.makedirs(save_dir, exist_ok=exist_ok)

        self.model = self.model.eval()
        if device is None:
            device = self.model.device

        if ds is not None:
            ds = self.prepare_ds(ds)
        else:
            ds = self.ds

        duplicate_map = defaultdict(list)
        non_duplicates = list()
        for obj in tqdm(ds):
            inputs = {}
            for name in self.feature_columns:
                inputs[name] = torch.LongTensor([obj[name]]).to(device)
            outputs = self.model.generate(
                **inputs,
                output_scores=True,
                return_dict_in_generate=True,
                **generate_kwargs)
            predict_ids = outputs.sequences[0]
            gen_label = self.tokenizer.decode(predict_ids, skip_special_tokens=True)
            gen_label = ' '.join(gen_label.split()[:max_label_words])
            scores = torch.nn.functional.softmax(outputs.scores[0], dim=-1)
            raw_obj = {k: v for k, v in obj.items() if k not in self.feature_columns}
            if gen_label == obj['label_text']:
                if threshold is not None and any(scores[0, i] for i in predict_ids) < threshold:
                    continue
                duplicate_map[obj['label_text']].append(raw_obj)
            else:
                non_duplicates.append(raw_obj)

        duplicate_total = 0
        nonduplicate_total = 0
        duplicate_path = os.path.join(save_dir, 'duplicate.jsonl')
        nonduplicate_path = os.path.join(save_dir, 'nonduplicate.jsonl')
        with open(duplicate_path, 'w') as writer, open(nonduplicate_path, 'w') as writer2:
            for obj in non_duplicates:
                writer2.writelines(json.dumps(obj, ensure_ascii=False) + '\n')
                nonduplicate_total += 1

            for label, obj_list in duplicate_map.items():
                writer2.writelines(json.dumps(obj_list[0], ensure_ascii=False) + '\n')
                nonduplicate_total += 1
                for obj in obj_list[1:]:
                    duplicate_total += 1
                    writer.writelines(json.dumps(obj, ensure_ascii=False) + '\n')
           
        print(f'{duplicate_total} duplicate text detected!')
        print(f'{nonduplicate_total} nonduplicate text detected!')
        print(f'Duplicate text has been saved to {duplicate_path}')
        print(f'Nonduplicate text has been saved to {nonduplicate_path}')



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="gen-dedup",
    version='0.0.1',
    author="Sean Lee",
    author_email="xmlee97@gmail.com",
    description="Generative deduplication",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/4AI/generative_deduplication",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
    install_requires=[
        "torch",
        "tqdm",
        "datasets",
        "transformers",
    ],
)
